# This file is part of Build Your Own CI
#
# Copyright 2018 Vincent Ladeuil
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
import os

from byoci import options
from byoci.slave import config
from byoci.tests import features
from byov import subprocesses
from byov.tests import features as vms_features


HERE = os.path.abspath(os.path.dirname(__file__))


ssh_agent_feature = features.ExecutableFeature('ssh-agent')


class LaunchpadToken(features.Feature):
    """Credentials for launchpad."""

    def __init__(self, lp_root='production'):
        super().__init__()
        if features.on_slave.available():
            conf = config.VmStack('localhost')
            self.path = os.path.join(conf.expand_options('{{byoci}.secrets}'),
                                     'launchpad', lp_root)
        else:
            self.path = os.path.join(options.testing_secrets_path,
                                     'launchpad', lp_root)

    def _probe(self):
        # We only test that the credentials file exists here, not its validity.
        return os.path.exists(self.path)

    def feature_name(self):
        return 'An OAuth token in {} for qastaging launchpad'.format(self.path)


# Parametrizing tests to run against production and qastaging starts here:
# launchpad_token is a singleton. It shouldn't. There should be one for
# production and one for qastaging. And controlled by a user option in a
# .conf-tests file.
launchpad_token = LaunchpadToken()


class BzrIdentity(features.Feature):
    """A feature capturing the current bzr whoami."""

    def __init__(self):
        super().__init__()
        self.identify = None

    def _probe(self):
        try:
            # 'cd ~' is needed for the edge case where the current directory is
            # a branch in a mounted file system that doesn't include the
            # associated (shared) repository. This happens for the
            # brz-slave-testing container when running the slave tests.  There
            # is no other use case that is known to make 'bzr whoami' fail
            # complaining that there is no repository around ;)
            _, out, _ = subprocesses.run(['sh', '-c', 'cd ~ && bzr whoami'])
            self.identity = out.strip()
            return True
        except subprocesses.errors.CommandError:
            return False

    def feature_name(self):
        return 'An existing bazaar whoami'


bzr_identity = BzrIdentity()


class GitIdentity(features.Feature):
    """A feature capturing the current git user email."""

    def __init__(self):
        super().__init__()
        self.name = None
        self.email = None

    def _probe(self):
        try:
            _, out, _ = subprocesses.run(['git', 'config', 'user.name'])
            self.name = out.strip()
            _, out, _ = subprocesses.run(['git', 'config', 'user.email'])
            self.email = out.strip()
            return True
        except subprocesses.errors.CommandError:
            return False

    def feature_name(self):
        return 'An existing git user name and email'


git_identity = GitIdentity()


class LaunchpadIdentity(features.Feature):
    """A feature capturing the current bzr launchpad login."""

    def __init__(self):
        super().__init__()
        self.identify = None

    def _probe(self):
        try:
            _, out, _ = subprocesses.run(['bzr', 'launchpad-login'])
            self.identity = out.strip()
            return True
        except subprocesses.errors.CommandError:
            return False

    def feature_name(self):
        return 'An existing launchpad login'


launchpad_identity = LaunchpadIdentity()


# Features relying on an environment property should be checked before any test
# isolation is setup. Once this is done, they can be freely used to decorate
# tests [classes] or during setUp().

launchpad_token.available()
bzr_identity.available()
git_identity.available()
launchpad_identity.available()
ssh_agent_feature.available()


# Useful shortcuts to export but not use internally
requires_existing_vm = vms_features.requires_existing_vm
ssh_feature = vms_features.ssh_feature
tests_config = vms_features.tests_config
requires = features.requires
test_requires = features.test_requires
