# This file is part of Build Your Own CI
#
# Copyright 2018 Vincent Ladeuil
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import time


from byoci.slave import (
    config,
    launchpad,
)
from byoci.tests import fixtures
from byoci.tests.slave import features
from byoci.tests.slave.branches import (
    bzr,
    git,
)
from byov import (
    config as vms_config,
    subprocesses,
    timeouts,
)


# Useful shortcuts
override_logging = fixtures.override_logging
patch = fixtures.patch


# Make sure user (or system) defined environment variables do not leak in the
# test environment.
isolated_environ = {
    'HOME': None,
    'BZR_PROGRESS_BAR': None,
}


# The project used for lp tests. See 'doc/tests.rst' for the manual (and
# required) steps to bootstrap.
launchpad_test_project_name = 'byoci-test-project'


def isolate_from_disk_for_slave(test):
    """Provide an isolated disk-based environment.

    A $HOME directory is setup so tests can setup arbitrary files including
    config ones.
    """
    fixtures.set_uniq_cwd(test)
    # Share the lxd certificate. We could generate one on the fly but assume
    # instead that one is already available since we require the lxc client
    # already.
    lxd_conf_dir = os.path.expanduser('~/.config/lxc')
    # Preserve some options from the outer environment
    outer = config.VmStack(None)
    setup_known_hosts_path = outer.get('byoci.setup_known_hosts')
    lp_login = outer.get('launchpad.login')
    ssh_key = outer.get('ssh.key')
    # Isolate tests from the user environment
    fixtures.isolate_from_env(test, isolated_environ)
    test.home_dir = os.path.join(test.uniq_dir, 'home')
    os.mkdir(test.home_dir)
    os.environ['HOME'] = test.home_dir
    os.environ['BZR_HOME'] = test.home_dir
    os.environ['LXD_CONF'] = lxd_conf_dir
    inner = config.VmStack(None)
    inner.set('byoci', 'selftest')
    # Inject preserved options
    inner.set('byoci.setup_known_hosts', setup_known_hosts_path)
    inner.set('launchpad.login', lp_login)
    inner.set('ssh.key', ssh_key)
    # Also isolate from the system environment
    test.etc_dir = os.path.join(test.uniq_dir,
                                vms_config.system_config_dir()[1:])
    os.makedirs(test.etc_dir)
    fixtures.patch(test, vms_config, 'system_config_dir', lambda: test.etc_dir)
    # Override some options for test purposes (generally the defaults for all
    # byoci workers and as such needing some care for tests to not stomp on
    # each other feet). Also, all expected uses of containers start in a
    # working directory below ~/workspace so mimic that behavior.
    test.workspace = os.path.join(test.home_dir, 'workspace')
    os.mkdir(test.workspace)
    os.chdir(test.workspace)
    return outer, inner


def define_uniq_container(test):
    """To isolate tests from each other, created containers need a unique name.

    To keep those names legal and still user-readable we use the class name and
    the test method name. The process id is added so that the same test suite
    can be run on the same host in several processes.
    """
    container_name = '{}-{}-{}'.format(test.__class__.__name__,
                                       test._testMethodName,
                                       os.getpid())
    # '_' are not valid in hostnames
    test.container_name = container_name.replace('_', '-')
    conf = config.VmStack(test.container_name)
    conf.set('vm.class', 'lxd')
    conf.set('vm.update', 'False')
    conf.set('vm.distribution', 'ubuntu')
    conf.set('vm.release', 'xenial')
    conf.set('vm.architecture', 'amd64')
    conf.store.save()
    conf.store.unload()


def use_launchpad(test, lp_root='production'):
    """Setup environment to use launchpad API and friends."""
    features.test_requires(test, features.launchpad_token)
    # Get original path before isolation
    credentials_path = features.launchpad_token.path
    isolate_from_disk_for_slave(test)
    # Setup to use launchpad API and friends
    conf = config.VmStack(None)
    conf.set('launchpad.service_root', lp_root)
    conf.set('launchpad.credentials', '~/.config/launchpad/{}'.format(lp_root))
    conf.store.save()
    conf.store.unload()
    os.mkdir(os.path.join(test.home_dir, '.config', 'launchpad'))
    with open(credentials_path) as fin:
        with open(conf.get('launchpad.credentials'), 'w') as fout:
            fout.write(fin.read())
    return launchpad.Launchpad(config.VmStack(None))


def set_bzr_identity(test):
    """Set the bzr whoami."""
    features.test_requires(test, features.bzr_identity)
    test.assertTrue(test.uniq_dir in os.environ['HOME'])
    subprocesses.run(['bzr', 'whoami', features.bzr_identity.identity])


def set_bzr_launchpad_access(test, lp_root):
    """Set the bzr launchpad login."""
    features.test_requires(test, features.launchpad_identity)
    test.assertTrue(test.uniq_dir in os.environ['HOME'])
    if lp_root == 'qastaging':
        # All tests run against qastaging
        plugin_dir = os.path.join(test.uniq_dir, 'home', '.bazaar', 'plugins')
        os.makedirs(plugin_dir)
        with open(os.path.join(plugin_dir, 'qastagingonly.py'), 'w') as f:
            f.write('from bzrlib.plugins.launchpad import lp_directory\n')
            f.write('lp_directory.LaunchpadService.DEFAULT_INSTANCE'
                    ' = "{}"'.format(lp_root))
    subprocesses.run(['bzr', 'launchpad-login',
                      features.launchpad_identity.identity])


def set_git_identity(test):
    """Set the git user name and email."""
    features.test_requires(test, features.git_identity)
    test.assertTrue(test.uniq_dir in os.environ['HOME'])
    subprocesses.run(['git', 'config', '--global', 'user.name',
                      features.git_identity.name])
    subprocesses.run(['git', 'config', '--global', 'user.email',
                      features.git_identity.email])


def set_git_launchpad_access(test, lp_root):
    """Configure git to access launchpad."""
    features.test_requires(test, features.launchpad_identity)
    test.assertTrue(test.uniq_dir in os.environ['HOME'])
    if lp_root == 'production':
        git_host = 'git.launchpad.net'
    elif lp_root == 'qastaging':
        git_host = 'git.qastaging.paddev.net'
    subprocesses.run(
        ['git', 'config', '--global',
         'url.ssh://{}@{}/.insteadOf'.format(
             features.launchpad_identity.identity,
             git_host),
         'lp:'])
    subprocesses.run(
        ['git', 'config', '--global', 'push.default', 'simple'])


class VcsBranchHandler(object):
    """Abstract base class to create local branches."""

    # Must be set by daughter classes
    factory = None
    """The class responsible for handling branches (or repositories)."""
    vcs_dir = None
    """A directory created by the vcs for storing its files."""

    def set_identity(self, test):
        """Setup the vcs identity."""
        raise NotImplementedError(self.set_identify)

    def set_launchpad_access(self, test, lp_root='production'):
        """Setup launchpad access for vcs."""
        raise NotImplementedError(self.set_launchpad_access)

    def lp_branch_url(self, stem):
        """Create a full launchpad url from a stem.

        This should embed the process pid so tests can be run concurrently
        without collision. A single (dedicated but arbitrary) project
        (launchpad_test_project_name) should be used to keep things isolated
        but simple. The MP url is unique as long as a single user doesn't
        attempt to run the test suite from different hosts at the same time and
        being unlucky enough to collide on the process id.

        """
        raise NotImplementedError(self.lp_test_url)

    def get_branch_from_lp(self, lp, url):
        """Get the launchpad object associated with the branch."""
        raise NotImplementedError(self.set_identify)


class BzrBranchHandler(VcsBranchHandler):

    factory = bzr.Branch
    vcs_dir = '.bzr'

    def set_identity(self, test):
        set_bzr_identity(test)

    def set_launchpad_access(self, test, lp_root='production'):
        set_bzr_launchpad_access(test, lp_root)

    def lp_branch_url(self, stem):
        return 'lp:~{}/{}/{}-{}'.format(
            features.launchpad_identity.identity,
            launchpad_test_project_name,
            stem,
            os.getpid())

    def get_branch_from_lp(self, lp, url):
        return lp.get_bzr_branch(url)


class GitRepositoryHandler(VcsBranchHandler):

    factory = git.Repository
    vcs_dir = '.git'

    def set_identity(self, test):
        set_git_identity(test)

    def set_launchpad_access(self, test, lp_root='production'):
        set_git_launchpad_access(test, lp_root)

    def lp_branch_url(self, stem):
        return 'lp:~{}/{}/+git/{}-{}'.format(
            features.launchpad_identity.identity,
            launchpad_test_project_name,
            stem,
            os.getpid())

    def get_branch_from_lp(self, lp, url):
        return lp.get_git_repo(url)


class VcsProposalHandler(object):
    """Abstract base class to interact with proposals."""

    def create(self, test, source, target, message, cover=None):
        """Create a merge proposal of source into target."""
        raise NotImplementedError(self.create_proposal)

    def build_comment_subject(self, proposal):
        """Build a standard subject for proposal comment."""
        raise NotImplementedError(self.build_comment_subject)


class BzrProposalHandler(VcsProposalHandler):

    def create(self, test, proposed, target, message, cover=None):
        if cover is None:
            cover = message
        return proposed.createMergeProposal(
            target_branch=target,
            initial_comment=cover,
            commit_message=message,
            needs_review=True)

    # FIXME: duplicated in LandApprovedBzrProposal.proposal_comment_subject
    # -- vila 2017-01-25
    def build_comment_subject(self, proposal):
        # Uses display_name (lp:xxx)
        return 'Re: [Merge] {} into {}'.format(
            proposal.source_branch.display_name,
            proposal.target_branch.display_name)


class GitProposalHandler(VcsProposalHandler):

    def get_ref(self, test, repo, branch):
        ref = repo.getRefByPath(path=branch)
        # Wait before giving up, this should be enough to cover
        # small maintenance windows or significant slowdowns.
        sleeps = timeouts.ExponentialBackoff(2, 600, 12)
        for sleep in sleeps:
            if ref is not None:
                break
            time.sleep(sleep)
            ref = repo.getRefByPath(path='master')
        try:
            test.assertIsNot(None, ref,
                             'Branch {} cannot be found inside {}'.format(
                                 branch, repo.unique_name))
        except AssertionError:
            # FIXME: Strictly speaking this should be an expected failure but
            # that will do for now -- vila 2018-01-17
            test.skipTest('launchpad git backend is broken')
        return ref

    def create(self, test, proposed, target, message, cover=None):
        if cover is None:
            cover = message
        proposed_git_ref = self.get_ref(test, proposed, 'master')
        test.assertEqual('refs/heads/master', proposed_git_ref.path)
        target_git_ref = self.get_ref(test, target, 'master')
        test.assertEqual('refs/heads/master', target_git_ref.path)
        return proposed_git_ref.createMergeProposal(
            merge_target=target_git_ref,
            initial_comment=cover,
            commit_message=message,
            needs_review=True)

    # FIXME: duplicated in LandApprovedGitProposal.proposal_comment_subject
    # -- vila 2017-01-25
    def build_comment_subject(self, proposal):
        # Uses unique_name (xxx)
        proposed_url = '{}:{}'.format(
            proposal.source_git_repository.unique_name,
            proposal.source_git_path[len('refs/heads/'):])
        target_url = '{}:{}'.format(
            proposal.target_git_repository.unique_name,
            proposal.target_git_path[len('refs/heads/'):])
        return 'Re: [Merge] {} into {}'.format(
            proposed_url,
            target_url)


def setup_proposal_target(test):
    """Create a local target branch to build proposals."""
    branch = test.branch_handler.factory('target')
    test.addCleanup(branch.delete)
    branch.create()
    branch.update('Something', 'file: foo\nfoo content\n')
    test.target_url = test.branch_handler.lp_branch_url('target')
    branch.push(test.target_url)
    test.target_lp_branch = test.branch_handler.get_branch_from_lp(
        test.lp, test.target_url)
    test.assertIsNot(None, test.target_lp_branch)
    test.addCleanup(test.target_lp_branch.lp_delete)


def create_proposal(test, name, approved=False, commit_msg=None, cover=None):
    """Create a merge proposal from a branch.

    :param test: A test case with a 'proposal_handler', 'branch_handler',
        'target_url' and 'target_lp_branch' attributes. A 'proposed_lp_branch'
        attribute will be created. See setup_proposal_target() to create
        'target_url' and 'target_lp_branch'.

    :param name: The branch name.

    :param approved: Should the proposal be approved.

    :param commit_msg: Commit message for the proposal. A default one is
        created is None is provided.

    :param cover: The cover letter for the proposal. Defaults to commit message
        if None is provided.

    """
    proposed = test.branch_handler.factory(name)
    # Strictly speaking we should create from the lp branch but we just
    # pushed 'target' there so this avoids a useless roundtrip.
    proposed.create_from('target')
    vcs_commit_msg = '{} purpose'.format(name)
    if commit_msg is None:
        commit_msg = vcs_commit_msg
    if cover is None:
        cover = commit_msg
    proposed.update(vcs_commit_msg,
                    'file: {}\n{} content\n'.format(name, name))
    proposed_lp_url = test.branch_handler.lp_branch_url(name)
    proposed.push(proposed_lp_url)
    test.proposed_lp_branch = test.branch_handler.get_branch_from_lp(
        test.lp, proposed_lp_url)
    test.addCleanup(test.proposed_lp_branch.lp_delete)
    proposal = test.proposal_handler.create(
        test, test.proposed_lp_branch, test.target_lp_branch,
        commit_msg, cover)
    if approved:
        subject = test.proposal_handler.build_comment_subject(proposal)
        proposal.createComment(subject=subject, vote='Approve',
                               content='Well done!')
        proposed_rev_id = proposed.tip_commit_id()
        proposal.setStatus(status='Approved', revid=proposed_rev_id)
    return proposal
